package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.tasks.FetchAllSlicesInfoTask;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.RecoverSliceTask;
import be.iminds.ilabt.jfed.experimenter_gui.tasks.TasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.property.SimpleStringProperty;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Stage;
import javafx.util.Callback;
import org.controlsfx.control.ButtonBar;
import org.controlsfx.control.action.Action;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import java.time.Duration;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;

/**
 * User: twalcari
 * Date: 12/5/13
 * Time: 9:21 PM
 */
public class RecoverSliceDialog extends VBox {
    private static final DateTimeFormatter EXPIRATION_TIME_FORMATTER = DateTimeFormatter.ofPattern("YYYY-MM-dd HH:mm");

    @FXML
    private TableView<Slice> slicesTableView;

    @FXML
    private TableColumn<Slice, String> sliceNameTableColumn;
    @FXML
    private TableColumn<Slice, String> subAuthTableColumn;
    @FXML
    private TableColumn<Slice, Instant> expirationTableColumn;
    @FXML
    private HBox initializingBox;

    @FXML
    private ProgressIndicator refreshProgressIndicator;
    @FXML
    private Label refreshLabel;

    private final EasyModel easyModel;
    private final HighLevelController hlc;
    private final TasksFactory tasksFactory;

    private final EventHandler<MouseEvent> mouseEventEventHandler = mouseEvent -> {
        if (mouseEvent.getClickCount() == 2) {
            recoverExperiment();
        }

    };

    private final Dialog dialog;

    private final TaskService taskService;
    private final Provider<Stage> parentStageProvider;

    @Inject
    RecoverSliceDialog(EasyModel easyModel, HighLevelController hlc, TasksFactory tasksFactory, TaskService taskService, @Named("parentStage") Provider<Stage> parentStageProvider) {
        this.easyModel = easyModel;
        this.hlc = hlc;
        this.tasksFactory = tasksFactory;
        this.taskService = taskService;
        this.parentStageProvider = parentStageProvider;

        //initialize the Scene
        FXMLUtil.injectFXML(this);

        slicesTableView.itemsProperty().bind(this.easyModel.slicesProperty());

        sliceNameTableColumn.setCellValueFactory(new PropertyValueFactory<>("name"));
        sliceNameTableColumn.setCellFactory(new Callback<TableColumn<Slice, String>, TableCell<Slice, String>>() {
            @Override
            public TableCell<Slice, String> call(TableColumn<Slice, String> sliceStringTableColumn) {
                return new TableCell<Slice, String>() {
                    {
                        setOnMouseClicked(mouseEventEventHandler);
                    }

                    @Override
                    protected void updateItem(String s, boolean b) {
                        if (s != null) {
                            setText(s);
                        } else {
                            setText("");
                        }
                    }
                };
            }
        });
        subAuthTableColumn.setCellValueFactory(t -> new SimpleStringProperty(t.getValue().getSubAuthorityName()));
        subAuthTableColumn.setCellFactory(new Callback<TableColumn<Slice, String>, TableCell<Slice, String>>() {
            @Override
            public TableCell<Slice, String> call(TableColumn<Slice, String> sliceStringTableColumn) {
                return new TableCell<Slice, String>() {
                    {
                        setOnMouseClicked(mouseEventEventHandler);
                    }

                    @Override
                    protected void updateItem(String s, boolean b) {
                        if (s != null) {
                            setText(s);
                        } else {
                            setText("");
                        }
                    }
                };
            }
        });

        expirationTableColumn.setCellValueFactory(param -> param.getValue().expirationDateProperty());
        expirationTableColumn.setCellFactory(new Callback<TableColumn<Slice, Instant>, TableCell<Slice, Instant>>() {
            @Override
            public TableCell<Slice, Instant> call(TableColumn<Slice, Instant> param) {
                return new TableCell<Slice, Instant>() {
                    @Override
                    protected void updateItem(Instant item, boolean empty) {
                        super.updateItem(item, empty);

                        if (item != null) {
                            setText(EXPIRATION_TIME_FORMATTER.format(item.atZone(ZoneId.systemDefault())));

                            Duration duration = Duration.between(Instant.now(),item);

                            if (duration.isNegative()) {
                                setTextFill(Color.RED);
                            } else if (duration.toHours() < 12) {
                                setTextFill(Color.ORANGERED);
                            } else if (duration.toHours() < 24) {
                                setTextFill(Color.ORANGE);
                            } else if (duration.toDays() < 7) {
                                setTextFill(Color.GREEN);
                            } else {
                                setTextFill(Color.DARKGREEN);
                            }
                        } else {
                            setText("");
                        }
                    }
                };
            }
        });

        //check if we have expiration dates to show
        boolean hasExpirationDate =
                easyModel.getSlices().stream().map(Slice::getExpirationDate).anyMatch(date -> date != null);
        updateColumns(hasExpirationDate);


        //sort by default on slice name
        sliceNameTableColumn.setComparator(String.CASE_INSENSITIVE_ORDER);
        slicesTableView.getSortOrder().add(sliceNameTableColumn);

        //show loading progressview when initializing
        javafx.concurrent.Task loadingTask = new javafx.concurrent.Task<Void>() {
            Task resolveTask = hlc.resolveUserAndSlices();

            @Override
            protected Void call() throws Exception {
                while (resolveTask.isActive()) {
                    Thread.sleep(100);
                }

                //check if some resolveSlice-calls are still running
                for (Slice slice : RecoverSliceDialog.this.easyModel.getSlices()) {
                    while (hlc.getAggregatesForSlice(slice).isActive())
                        Thread.sleep(100);
                }


                return null;
            }

            @Override
            protected void succeeded() {
                initializingBox.setVisible(false);
            }
        };

        final Thread thread = new Thread(loadingTask);
        thread.setDaemon(true);
        thread.setName("Recover-Loading-Task");
        thread.start();

        dialog = new Dialog(parentStageProvider.get(), "Recover an existing experiment", false /* lightweight */);

        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.AMBULANCE, Color.GREEN));
        dialog.setMasthead("Select the experiment to recover");
        dialog.setContent(this);

        Action refreshAction = new DialogAction("Refresh & Fetch exp. dates", ButtonBar.ButtonType.OTHER,
                false, false, false, actionEvent -> refreshSlicesList());
        refreshAction.setGraphic(GlyphUtils.createSmallGlyph(FontAwesome.Glyph.REFRESH, Color.GREEN).useGradientEffect());

        Action startExperimentAction =
                new DialogAction("Recover Experiment", ButtonBar.ButtonType.OK_DONE, false, false, true,
                        actionEvent -> recoverExperiment());
        dialog.getActions().setAll(refreshAction, startExperimentAction);

    }

    private void updateColumns(boolean showExpirationColumn) {
        sliceNameTableColumn.prefWidthProperty().unbind();
        subAuthTableColumn.prefWidthProperty().unbind();
        expirationTableColumn.prefWidthProperty().unbind();

        expirationTableColumn.setVisible(showExpirationColumn);


        sliceNameTableColumn.prefWidthProperty().bind(slicesTableView.widthProperty().multiply(
                showExpirationColumn ? 0.4 : 0.6).subtract(1));
        subAuthTableColumn.prefWidthProperty().bind(slicesTableView.widthProperty().multiply(
                showExpirationColumn ? 0.25 : 0.4).subtract(1));
        expirationTableColumn.prefWidthProperty().bind(slicesTableView.widthProperty().multiply(
                showExpirationColumn ? 0.35 : 0.0).subtract(1));
    }


    public void showDialog() {
        dialog.show();
    }

    private void refreshSlicesList() {
        refreshProgressIndicator.visibleProperty().unbind();
        refreshLabel.textProperty().unbind();

        FetchAllSlicesInfoTask fetchAllSlicesInfoTask = tasksFactory.createFetchAllSlicesInfoTask();

        refreshProgressIndicator.visibleProperty().bind(fetchAllSlicesInfoTask.runningProperty());
        refreshLabel.textProperty().bind(fetchAllSlicesInfoTask.messageProperty());

        fetchAllSlicesInfoTask.setOnSucceeded(event -> {
            boolean hasExpirationDate =
                    easyModel.getSlices().stream().anyMatch(slice -> slice.getExpirationDate() != null);

            updateColumns(hasExpirationDate);
        });


        taskService.submitTask(fetchAllSlicesInfoTask);

    }

    public void recoverExperiment() {

        if (slicesTableView.getSelectionModel().isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("Please select the slice you wish to recover")
                    .masthead("No slice selected")
                    .showError();
            return;
        }
        dialog.hide();

        final Slice recoverSlice = slicesTableView.getSelectionModel().getSelectedItem();

        RecoverSliceTask recoverSliceTask = tasksFactory.createRecoverSliceTask(recoverSlice);
        taskService.submitTask(recoverSliceTask);


    }
}
